
#
# Written as part of DPSA-project, by @MxmUrw, @ooovi
#
# Based on `dpfedavg_numpy_client.py`.
#

"""Wrapper for configuring a Flower client for usage with DPSA."""

from typing import Dict, Optional, Tuple, Any, List

import numpy as np

from flwr.client.numpy_client import NumPyClient
from flwr.common.typing import Config, NDArrays, NDArray, Scalar

from dpsa4fl_bindings import (
    client_api_new_state,
    client_api_get_privacy_parameter,
    client_api_submit,
)

from .reshaping_config import ReshapingConfig, readReshapingConfig, writeReshapingConfig



class DPSANumPyClient(NumPyClient):
    """
    A flower client for federated learning with global differential privacy and
    secure aggregation. Uses the dpsa project infrastructure, see here
    for more information: https://github.com/dpsa-project/overview

    NOTE: This is intended for use with the DPSAServer flower server.

    Attributes
    ----------
    privacy_spent: float
        Zero-concentrated Differential Privacy of this client's data
        spent since DPSANumPyClient object construction. Accumulates over
        training rounds.
    """

    def __init__(
        self,
        max_privacy_per_round: float,
        aggregator1_location: str,
        aggregator2_location: str,
        client: NumPyClient,
        allow_evaluate = False
    ) -> None:
        """
        Parameters
        ----------
        max_privacy_per_round: float
            The maximal zero-contentrated differential privacy budget allowed to
            be spent on a single round of training. If the selected server offers
            a weaker guarantee, no data will be submitted and an exception will be raised.
        aggregator1_location: str
            Location of the first aggregator server in URL format including the port.
            For example, for a server running locally: "http://127.0.0.1:9991"
        aggregator2_location: str
            Location of the second aggregator server in URL format including the port.
            For example, for a server running locally: "http://127.0.0.1:9992"
        client: NumPyClient
            The NumPyClient used for executing the local learning tasks.
        allow_evaluate: bool
            Evaluation is a privacy-relevant operation on the client dataset. If this flag
            is set to `False`, evaluation always reports infinite loss and zero accuracy to
            the server. Otherwise, the evaluation function of the wrapped client will be used
            and the results will be released to the server, potentially compromising privacy.
            Defaults to `False`.
        """
        super().__init__()
        self.max_privacy_per_round = max_privacy_per_round
        self.client = client
        self.dpsa4fl_client_state = client_api_new_state(
            aggregator1_location,
            aggregator2_location,
        )
        # self.shapes = None
        # self.split_indices = None
        self.privacy_spent = 0
        self.allow_evaluate = allow_evaluate

    def get_properties(self, config: Config) -> Dict[str, Scalar]:
        return self.client.get_properties(config)

    def get_parameters(self, config: Dict[str, Scalar]) -> NDArrays:
        return self.client.get_parameters(config)

    def reshape_parameters(self, parameters: NDArrays, reshaping_config: Optional[ReshapingConfig]) -> NDArrays:
        """
        Reshape parameters given as a (1,)-NDArray into the Client's model shape,
        as given in the `shapes` attribute. Used for recovering the shape of a
        gradient that originated from the DPSA aggregation server and was hence in
        single-vector shape.

        Parameters
        ----------
        parameters: NDArrays
            The parameters to be reshaped. Expected to be (1,)-shaped.
        reshaping_config: Optional[ReshapingConfig]
            The format to reshape the parameters into. Should be taken from the output of the
            `flatten_parameters` function. If value is `None`, no reshaping is done.

        Returns
        -------
        parameters: NDArrays
            The input parameters reshaped to match this Client's `shapes` attribute.
        """
        # update parameter shapes
        # if we are in first round (self.shapes is None), then we don't need to reshape.
        # But if we are in any following round, then we need to take our previous shapes
        # and lengths and reshape the `parameters` argument accordingly
        if reshaping_config is not None:
            shapes = reshaping_config.shapes
            split_indices = reshaping_config.split_indices

            assert len(split_indices) + 1 == len(shapes), "Expected #indices = #shapes - 1"

            print("In follow-up round, reshaping. length of params is: ", len(parameters))
            assert len(parameters) == 1, "Expected parameters to have length 1!"

            single_array = parameters[0]
            print("Found single ndarray of shape ", single_array.shape, " and size ", single_array.size)
            # assert single_array.shape == (,), "Wrong ndarray shape!"

            # debug: print highest element
            print("(reshape) highest nan element is: ", np.amax(single_array))
            print("(reshape) highest nonnan element is: ", np.nanmax(single_array))

            # split and reshape
            arrays = np.split(single_array, split_indices)
            print("After splitting, have ", len(arrays), " arrays")

            arrays = [np.reshape(a,s) for (a,s) in zip(arrays, shapes)]
            print("Now have the following shapes:")
            for a in arrays:
                print(a.shape)

            # check that we have our value at position 449
            rval1 = np.around(arrays[0], decimals = 2)
            print("in array at 449 have: ", rval1[0:1, 0:1, 0:1, 0:1])
            print(rval1)

            # change parameters to properly shaped list of arrays
            parameters = arrays

        else:
            print("In first round, not reshaping.")

        return parameters

    def flatten_parameters(self, params: NDArrays) -> Tuple[NDArray, ReshapingConfig]:
        """
        Reshape the input parameters into a (1,)-NDArray to prepare
        them for sumbission to the DPSA infrastructure.

        Parameters
        ----------
        params: NDArrays
            The parameters to be flattened. Flattening in row-major order.

        Returns
        -------
        flat_param_vector: NDArray
            The input parameters reshaped to one long (1,)-NDArray.

        reshaping_config: ReshapingConfig
            Format for reshaping the flat array back into its original shape.
        """

        # print param shapes
        print("The shapes are:")
        for p in params:
            print(p.shape)

        # flatten params before submitting
        shapes = [p.shape for p in params]
        flat_params = [p.flatten('C') for p in params]
        p_lengths = [p.size for p in flat_params]

        # loop
        # (convert p_lengths into indices because ndarray.split takes indices instead of lengths)
        split_indices = []
        current_index = 0
        for l in p_lengths:
            split_indices.append(current_index)
            current_index += l
        split_indices.pop(0) # need to throw away first element of list


        flat_param_vector = np.concatenate(flat_params)


        print("vector length is: ", flat_param_vector.shape)

        # debug: print highest element
        print("highest nan element is: ", np.amax(flat_param_vector))
        print("highest nonnan element is: ", np.nanmax(flat_param_vector))

        return flat_param_vector, ReshapingConfig(shapes, split_indices)

    def fit(
        self, params0: NDArrays, config: Dict[str, Scalar]
    ) -> Tuple[NDArrays, int, Dict[str, Scalar]]:
        """
        Train the provided parameters using the locally held dataset. After
        training locally using the wrapped NumPyClient, the gradients are
        clipped to norm 1 and submitted to the DPSA infrastructure for secure
        aggregation of all clients' gradients and noising for differential privacy.

        Parameters
        ----------
        params0 : NDArrays
            The current (global) model parameters.
        config : Dict[str, Scalar]
            Configuration parameters which allow the
            server to influence training on the client. It can be used to
            communicate arbitrary values from the server to the client, for
            example, to set the number of (local) training epochs.

        Returns
        -------
        parameters : NDArrays
            The locally updated model parameters.
        num_examples : int
            The number of examples used for training.
        metrics : Dict[str, Scalar]
            A dictionary mapping arbitrary string keys to values of type
            bool, bytes, float, int, or str. It can be used to communicate
            arbitrary values back to the server.
        """

        # get current task_id
        task_id = config['task_id']

        # get reshaping parameters
        reshaping_config = readReshapingConfig(config)

        #reshape params
        params0 = self.reshape_parameters(params0, reshaping_config)

        # train on data
        params, i, d = self.client.fit(params0, config)

        # compute gradient
        grad = [np.subtract(p, p0) for (p , p0) in zip(params,params0)]

        flat_grad_vector, new_reshaping_config = self.flatten_parameters(grad)
        writeReshapingConfig(d, new_reshaping_config)

        # truncate if norm >= 1
        norm = np.linalg.norm(flat_grad_vector)
        print("norm of vector is: ", norm)
        if norm >= 1:
            print("Need to scale vector")
            flat_grad_vector = flat_grad_vector * (1/(norm + 0.01))
            norm = np.linalg.norm(flat_grad_vector)
            print("now norm of vector is: ", norm)

        # get server privacy parameter
        eps = client_api_get_privacy_parameter(self.dpsa4fl_client_state, task_id)

        if eps > self.max_privacy_per_round:
            raise Exception("DPSAClient requested at least " + str(self.max_privacy_per_round) + " privacy but server supplied only " + str(eps))
        else:
            # log privacy loss
            self.privacy_spent += eps

            # submit data to janus
            client_api_submit(self.dpsa4fl_client_state, task_id, flat_grad_vector)

            # return empty, parameter update needs to be retrieved from janus
            return [], i, d

    def evaluate(
        self, parameters: NDArrays, config: Dict[str, Scalar]
    ) -> Tuple[float, int, Dict[str, Scalar]]:
        if self.allow_evaluate:
            reshaping_config = readReshapingConfig(config)
            parameters = self.reshape_parameters(parameters, reshaping_config)
            return self.client.evaluate(parameters, config)
        else:
            return float('inf'), 1, {"accuracy": 0}




"""Flower server."""


from logging import WARNING
from typing import Callable, Dict, List, Optional, Tuple, Union

from dpsa4fl_bindings import (
    controller_api_new_state,
    controller_api_create_session,
    controller_api_start_round,
    controller_api_collect,
    controller_api_end_session,
    PyControllerState
)

import numpy as np

from flwr.common import (
    EvaluateIns,
    EvaluateRes,
    FitIns,
    FitRes,
    MetricsAggregationFn,
    NDArrays,
    Parameters,
    Scalar,
    ndarrays_to_parameters,
    parameters_to_ndarrays,
)
from flwr.common.logger import configure, log
from flwr.server.client_manager import ClientManager
from flwr.server.client_proxy import ClientProxy
from flwr.server.server import Server, FitResultsAndFailures
from flwr.server.strategy.aggregate import aggregate, weighted_loss_avg
from flwr.server.strategy import Strategy, FedAvg

from .reshaping_config import ReshapingConfig, readReshapingConfig, writeReshapingConfig


class DPSAServer(Server):
    """
    A flower server for federated learning with global differential privacy and
    secure aggregation. Uses the dpsa project infrastructure, see here
    for more information: https://github.com/dpsa-project/overview

    NOTE: This is intended for use with the DPSANumPyClient flower client.
    """ 

    def __init__(
        self,
        *,
        model_size: int,
        privacy_parameter: float,
        granularity: int,
        aggregator1_location: str,
        aggregator2_location: str,
        client_manager: ClientManager,
        strategy: Optional[Strategy] = None
    ) -> None:
        """
        Parameters
        ----------
        model_size: int
            The number of parameters of the model to be trained.
        privacy_parameter: float
            The desired privacy per learning step. One aggregation step will
            be `1/2*privacy_parameter^2` zero-concentrated differentially private
            for each client.
        granularity: int
            The resolution of the fixed-point encoding used for secure aggregation.
            A larger value will result in a less lossy representation and more
            communication and computation overhead. Currently, 16, 32 and 64 bit are
            supported.
        aggregator1_location: str
            Location of the first aggregator server in URL format including the port.
            For example, for a server running locally: "http://127.0.0.1:9991"
        aggregator2_location: str
            Location of the second aggregator server in URL format including the port.
            For example, for a server running locally: "http://127.0.0.1:9992"
        """

        # call dpsa4fl to create state object
        #
        # NOTE: this might fail, e.g. if servers are not running at the given locations
        try:
            self.dpsa4fl_state = controller_api_new_state(
                model_size,
                privacy_parameter,
                granularity,
                aggregator1_location,
                aggregator2_location,
            )
        except RuntimeError as err:
            print("=======================================")
            print("Could not initialize controller state.")
            print(f"Are aggregator servers running at {aggregator1_location} and {aggregator2_location}?")
            print("=======================================")
            print("")
            print(f"The original error message is:")
            print("")
            print(err)

        dpsa4fl_strategy = DPSAStrategyWrapper(
            strategy = strategy if strategy is not None else FedAvg(),
            dpsa4fl_state = self.dpsa4fl_state
        )

        super().__init__(client_manager=client_manager, strategy=dpsa4fl_strategy)

        # call dpsa4fl to create new session
        controller_api_create_session(self.dpsa4fl_state)
        
    """End the dpsa4fl session. Use at the end of training for graceful shutdown."""
    def __del__(self):
        # end session when we are done
        controller_api_end_session(self.dpsa4fl_state)

    """Perform a single round of federated averaging."""
    def fit_round(
        self,
        server_round: int,
        timeout: Optional[float],
    ) -> Optional[
        Tuple[Optional[Parameters], Dict[str, Scalar], FitResultsAndFailures]
    ]:
        # Call dpsa4fl to start a new round
        controller_api_start_round(self.dpsa4fl_state)

        # The rest of the work is done by the `DPSAStrategyWrapper`
        # which is called in the server implementation of super.
        res = super().fit_round(server_round, timeout)

        return res


# flake8: noqa: E501
class DPSAStrategyWrapper(Strategy):
    """
    Configurable strategy implementation for federated learning with
    global differential privacy and secure aggregation. To be used together
    with the DPSAServer.
    """
    def __init__(self, *, strategy: Strategy, dpsa4fl_state: PyControllerState) -> None:
        """
        Parameters
        ----------
        strategy: Strategy
            The strategy to be wrapped. The wrapping replaces the `configure_fit` and
            `aggregate_fit` methods with ones that interact with the dpsa
            infrastructure. All other methods are copied from the input strategy.
        dpsa4fl_state: PyControllerState
            State object containing training parameters and server locations necessary
            for interoperation with dpsa. See `dpsa4fl_bindings.controller_api_new_state`
            for details.
        """ 
        super().__init__()
        self.strategy = strategy
        self.dpsa4fl_state = dpsa4fl_state
        self.parameters = Parameters(tensors=[], tensor_type="")
        self.reshaping_config : Optional[ReshapingConfig] = None

    def initialize_parameters(
        self, client_manager: ClientManager
    ) -> Optional[Parameters]:
        return self.strategy.initialize_parameters(client_manager)

    def configure_fit(
        self, server_round: int, parameters: Parameters, client_manager: ClientManager
    ) -> List[Tuple[ClientProxy, FitIns]]:
        result = self.strategy.configure_fit(server_round, parameters, client_manager)
        def append_config(arg: Tuple[ClientProxy, FitIns]):
            client, fitins = arg

            # add the task_id into the config
            fitins.config['task_id'] = self.dpsa4fl_state.mstate.task_id

            # add the reshaping config if available
            if self.reshaping_config is not None:
                writeReshapingConfig(fitins.config, self.reshaping_config)

            return (client, fitins)

        self.parameters = parameters

        return list(map(append_config, result))


    def aggregate_fit(
        self,
        server_round: int,
        results: List[Tuple[ClientProxy, FitRes]],
        failures: List[Union[Tuple[ClientProxy, FitRes], BaseException]],
    ) -> Tuple[Optional[Parameters], Dict[str, Scalar]]:

        # check that the parameter shapes of all clients match
        reshaping_configs = [res.metrics.get('reshaping_config') for (_, res) in results]
        print(f"reshaping configs are: {reshaping_configs}")

        # NOTE: We have to do the following conversion with the encoded configs
        #       because classes containing lists are not hashable by default
        #       and thus a list of them cannot be unique-ified by turning it into a set.
        unique_reshaping_configs = list(set(reshaping_configs))

        if len(unique_reshaping_configs) > 1:
            print("there have been different reshaping configs")
            exit(-1);
        else:
            # Here we use the first element of the original list
            self.reshaping_config = readReshapingConfig(results[0][1].metrics)

        # we do our custom aggregation here.
        print("Getting results from janus")
        collected: np.ndarray = controller_api_collect(self.dpsa4fl_state)
        print("Done getting results from janus, vector length is: ", collected.shape)

        # convert ndarray type
        flat_grad_array = collected.astype(np.float32)

        old_params_arrays = parameters_to_ndarrays(self.parameters)
        flat_old_params_array = old_params_arrays[0]

        # if old params is not flat, need to flatten
        if len(old_params_arrays) > 1:
            flat_old_params = [p.flatten('C') for p in old_params_arrays]
            flat_old_params_array = np.concatenate(flat_old_params)

        # add gradient to current params
        flat_param_array = flat_old_params_array + flat_grad_array

        # encode again in params format
        parameters_aggregated = ndarrays_to_parameters([flat_param_array])

        # Aggregate custom metrics if aggregation fn was provided
        # the same as what happens in the FedAvg strategy
        metrics_aggregated = {}
        if hasattr(self.strategy, 'fit_metrics_aggregation_fn'):
            if self.strategy.fit_metrics_aggregation_fn:
                fit_metrics = [(res.num_examples, res.metrics) for _, res in results]
                metrics_aggregated = self.strategy.fit_metrics_aggregation_fn(fit_metrics)
            elif server_round == 1:  # Only log this warning once
                log(WARNING, "No fit_metrics_aggregation_fn provided")
        elif server_round == 1:  # Only log this warning once
            log(WARNING, "Strategy has no fit_metrics_aggregation_fn")

        return parameters_aggregated, metrics_aggregated

    def configure_evaluate(
        self, server_round: int, parameters: Parameters, client_manager: ClientManager
    ) -> List[Tuple[ClientProxy, EvaluateIns]]:
        return self.strategy.configure_evaluate(server_round, parameters, client_manager)

    def aggregate_evaluate(
        self,
        server_round: int,
        results: List[Tuple[ClientProxy, EvaluateRes]],
        failures: List[Union[Tuple[ClientProxy, EvaluateRes], BaseException]],
    ) -> Tuple[Optional[float], Dict[str, Scalar]]:
        return self.strategy.aggregate_evaluate(server_round, results, failures)

    def evaluate(
        self, server_round: int, parameters: Parameters
    ) -> Optional[Tuple[float, Dict[str, Scalar]]]:
        return self.strategy.evaluate(server_round, parameters)





from dataclasses import dataclass
from typing import Dict, Optional, Tuple, List
from flwr.common.typing import Scalar
from dataclass_wizard import JSONWizard

Shapes = List[Tuple[int, ...]]
SplitIndices = List[int]

@dataclass
class ReshapingConfig(JSONWizard):
    shapes: Shapes
    split_indices: SplitIndices

def readReshapingConfig(d: Dict[str, Scalar]) -> Optional[ReshapingConfig]:
    reshaping_config_encoded = d.get('reshaping_config')

    if isinstance(reshaping_config_encoded, str):
        c = ReshapingConfig.from_json(reshaping_config_encoded)
        if isinstance(c, ReshapingConfig):
            return c
        else:
            print("WARNING: Got a list instead of a single object! (when deserializing a ReshapingConfig)")
            return None
    else:
        return None

def writeReshapingConfig(d: Dict[str, Scalar], c: ReshapingConfig) -> Dict[str, Scalar]:
    d['reshaping_config'] = c.to_json()
    return d




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dpsa4flower']

package_data = \
{'': ['*']}

install_requires = \
['dataclass-wizard>=0.22.2,<0.23.0',
 'dpsa4fl-bindings>=0.1.3,<0.2.0',
 'flwr>=1.4.0,<2.0.0',
 'numpy>=1.24.3,<2.0.0']

setup_kwargs = {
    'name': 'dpsa4flower',
    'version': '0.1.0',
    'description': 'A server and client to use the flower framework with DPSA.',
    'long_description': '# dpsa4flower\nServer and client to use the [flower framework](https://flower.dev/) for differentially private federated learning with secure aggregation.\n\nMade to be used with the [dpsa infrastructure](https://github.com/dpsa-project/overview), head there for an explanation of the system\'s participants and properties. Setup of additional aggregation servers is required, head [here](https://github.com/dpsa-project/dpsa4fl-testing-infrastructure) for instructions.\n\n## Installation\nTo install, you require the following packages:\n- python version 3.9 or higher.\n- [poetry](https://python-poetry.org/) package manager for python\n\nOnce you have those, go ahead and clone this repository:\n```\n> git clone https://github.com/dpsa-project/dpsa4flower.git\n```\nEnter the new directory:\n```\n> cd dpsa4flower\n```\nUse poetry to create a virtualenv and install all dependencies:\n```\n> poetry shell\n> poetry install\n```\nYou\'re ready to use our classes now. Note that to actually run a learning task, you will need to provide locations at which two seperate dpsa4fl aggregation servers are running. See [here](https://github.com/dpsa-project/dpsa4fl-testing-infrastructure) for instructions or check out our example project.\n\n## Example code\nThere is a [repo](https://github.com/dpsa-project/dpsa4fl-example-project) containing an example implementation learning the CIFAR task using a torch model, where learning is federated using flower with differential privacy and secure aggregation.\n\n## Classes\nThis package exposes two classes, one for the server and one for the client.\n### `DPSAServer(model_size, privacy_parameter, granularity, aggregator1_location, aggregator2_location, client_manager, strategy)`\n\nThe [dpsa4flower server class](https://github.com/dpsa-project/dpsa4flower/blob/3f1becb09bb79dfe26f9ee959114cf6c36a31dbb/dpsa_flower/dpsa_server.py#L40) extends the [flower server class](https://flower.dev/docs/apiref-flwr.html#module-flwr.server) with the necessities for using DPSA for aggregation. It handles configuration of the aggregator servers, reshaping of the collected aggregation results, and redistributing the updates to the clients. Construction requires the following parameters as keyword arguments:\n\n- `model_size: int` The number of parameters of the model to be trained.\n- `privacy_parameter: float` The desired privacy per learning step. One aggregation step will\n    be `1/2*privacy_parameter^2` zero-concentrated differentially private\n    for each client.\n- `granularity: int` The resolution of the fixed-point encoding used for secure aggregation.\n    A larger value will result in a less lossy representation and more\n    communication and computation overhead. Currently, 16, 32 and 64 bit are\n    supported.\n- `aggregator1_location: str` Location of the first aggregator server in URL format including the port.\n    For example, for a server running locally: "http://127.0.0.1:9991"\n- `aggregator2_location: str` Location of the second aggregator server in URL format including the port.\n    For example, for a server running locally: "http://127.0.0.1:9992"\n- `client_manager: flwr.server.ClientManager` A flower client manager to manage connected clients.\n- `strategy: Optional[flwr.server.strategy.Strategy]` A flower strategy for the server to use. It will be wrapped replacing the `configure_fit` and `aggregate_fit` methods with ones that interact with the dpsa infrastructure.\n\nAn example construction of a dpsa4flower server object would look like this:\n```python\ndpsa4flower.DPSAServer(\n        model_size = 62006,\n        privacy_parameter = 30,\n        granularity = 32,\n        aggregator1_location = "http://127.0.0.1:9981",\n        aggregator2_location = "http://127.0.0.1:9982",\n        client_manager=flwr.server.SimpleClientManager(),\n)\n```\nThe created object can then be used to start a flower server using [`flwr.server.start_server`](https://flower.dev/docs/apiref-flwr.html#server-start-server) as usual.\n\n### `DPSANumPyClient(max_privacy_per_round, aggregator1_location, aggregator2_location, client, allow_evaluate)`\nThe [dpsa4flower client class](https://github.com/dpsa-project/dpsa4flower/blob/3f1becb09bb79dfe26f9ee959114cf6c36a31dbb/dpsa_flower/dpsa_numpy_client.py#L19) implements the [`NumPyClient`](https://flower.dev/docs/apiref-flwr.html#numpyclient) interface provided by flower. It\'s a wrapper for existing `NumPyClient`s adding secure aggregation and differential privacy. The wrapped client is used for local training, results are then submitted to the secure aggregation infrastructure in an encrypted fashion. The constructor requires the following parameters as keyword arguments:\n \n- `max_privacy_per_round: float` The maximal zero-contentrated differential privacy budget allowed to be spent on a single round of training. If the selected server offers a weaker guarantee, no data will be submitted and an exception will be raised.\n- `aggregator1_location: str` Location of the first aggregator server in URL format including the port. For example, for a server running locally: "http://127.0.0.1:9991"\n- `aggregator2_location: str` Location of the second aggregator server in URL format including the port. For example, for a server running locally: "http://127.0.0.1:9992"\n- `client: flower.client.numpy_client.NumPyClient` The NumPyClient used for executing the local learning tasks.\n- `allow_evaluate: bool` Evaluation is a privacy-relevant operation on the client dataset. If this flag is set to `False`, evaluation always reports infinite loss and zero accuracy to the server. Otherwise, the evaluation function of the wrapped client will be used and the results will be released to the server, potentially compromising privacy. Defaults to `False`.\n\nAn example construction of a dpsa4flower client object would look like this:\n```python\ndpsa4flower.DPSANumPyClient(\n    max_privacy_per_round = 30,\n    aggregator1_location = "http://127.0.0.1:9981",\n    aggregator2_location = "http://127.0.0.1:9982",\n    client = FlowerClient()\n)\n```\nwhere `FlowerClient` is some [`NumPyClient`](https://flower.dev/docs/apiref-flwr.html#numpyclient) of your choice. It can then be started using [`flwr.client.start_numpy_client`](https://flower.dev/docs/apiref-flwr.html#flwr.client.start_numpy_client) as usual.\n\n## What\'s going on\nWhen using our classes in the setup described (and used in the example project), the training procedure takes place as described in this diagram:\n\n\n```\n                                             gradient sum\n          ┌───────────────────────────────────────────────────────────────────────────────┐\n          │                             (differentially private)                          │\n          │                                                                               │\n          │                                                                               │\n          │                                                                               │\n          │                                                                               │\n          │             gradient shares                                                   │\n          │              (ciphertext)                                                     │\n  ┌───────▼─────────┬────────────────────┐                                                │\n  │ DPSANumPyClient │                    │ ┌──────────────┐                               │\n  └─────────────────┴────────────────┐   └─►              │                               │\n                                     │     │ Aggregator 1 ├───┐                           │\n                                 ┌───)─────►              │   │                           │\n                                 │   │     └──────────────┘   │                   ┌───────┴───────┐\n          .                      │   │                        │    gradient sum   │               │\n          .                      │   │                        ├───────────────────►  DPSAServer   │\n          .                      │   │                        │  (differentially  │               │\n                                 │   │     ┌──────────────┐   │      private)     └───────────────┘\n                                 │   └─────►              │   │\n                                 │         │ Aggregator 2 ├───┘\n  ┌─────────────────┬────────────┘   ┌─────►              │\n  │ DPSANumPyClient │                │     └──────────────┘\n  └─────────────────┴────────────────┘\n                     gradient shares\n                      (ciphertext)\n\n\n\n     flower clients                     dpsa4fl infrastructure                     flower server\n     --------------                     ----------------------                     -------------\ncompute gradients locally          checks if clipping was done properly,        collects aggregate,\n   on sensitive data,               computes aggregate on ciphertext,           distributes updates\nclip to norm 1 and submit          adds noise for differential privacy.         back to the clients\n                                     ciphertext can not be decrypted\n                                     if the servers don\'t collaborate\n```\n',
    'author': 'Maxim Urschumzew',
    'author_email': 'u.maxim@live.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
